import sys
from tqdm.auto import tqdm
import os
import pickle
import urllib
import json
import http
import numpy as np
import time
# from google.cloud import language
# from google.cloud.language import enums
# from google.cloud.language import types
from google.cloud import language_v1
import boto3
def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

class OnlinePredictor:
    def __init__(self, pred_file, batch_size=1000, wait_time=1, model='azure3'):
        self.preds = {}
        self.post_fn = None
        if os.path.exists(pred_file):
            self.preds = pickle.load(open(pred_file, 'rb'))
        self.pred_file = pred_file
        if model == 'azure3':
            self.pp_fn = self.predict_proba_azure3
        if model == 'azure2':
            self.pp_fn = self.predict_proba_azure2
            self.post_fn = self.binary_to_three
        if model == 'google':
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = '/home/slundberg/.google_cloud_key.json'
            self.pp_fn = self.predict_proba_google
            self.post_fn = self.binary_to_three
            # self.gclient = language.LanguageServiceClient()
            self.gclient = language_v1.LanguageServiceClient()
        if model == 'amazon':
            self.pp_fn = self.predict_proba_amazon
            self.client = boto3.client('comprehend', region_name='us-west-2')

        self.batch_size = batch_size
        self.wait_time = wait_time

    def binary_to_three(self, pr):
        # This is what google recommends
        margin_neutral = 0.25
        mn = margin_neutral / 2.
        pp = np.zeros((pr.shape[0], 3))
        neg = pr < 0.5 - mn
        pp[neg, 0] = 1 - pr[neg]
        pp[neg, 2] = pr[neg]
        pos = pr > 0.5 + mn
        pp[pos, 0] = 1 - pr[pos]
        pp[pos, 2] = pr[pos]
        neutral_pos = (pr >= 0.5) * (pr < 0.5 + mn)
        pp[neutral_pos, 1] = 1 - (1 / margin_neutral) * np.abs(pr[neutral_pos] - 0.5)
        pp[neutral_pos, 2] = 1 - pp[neutral_pos, 1]
        neutral_neg = (pr < 0.5) * (pr > 0.5 - mn)
        pp[neutral_neg, 1] = 1 - (1 / margin_neutral) * np.abs(pr[neutral_neg] - 0.5)
        pp[neutral_neg, 0] = 1 - pp[neutral_neg, 1]
        return (pp.T / pp.sum(1)).T # normalize to sum to 1

    def predict_proba(self, exs, silent=True):
        to_pred = [x for x in exs if x not in self.preds]
        chunked = list(chunks(to_pred, self.batch_size))
        for docs in chunked:
            pps = self.pp_fn(docs)
            for x, pp in zip(docs, pps):
                self.preds[x] = pp
            pickle.dump(self.preds, open(self.pred_file, 'wb'))
            time.sleep(self.wait_time)
        ret = np.array([self.preds.get(x) for x in exs])
        if self.post_fn:
            ret = self.post_fn(ret)
        return ret

    def predict_and_confidences(self, exs):
        confs = self.predict_proba(exs)
        preds = np.argmax(confs, axis=1)
        return preds, confs


    def predict_proba_amazon(self, exs):
#         print('amazon: predicting %d examples' % len(exs))
        scores = []
        for text in exs:
            ret = self.client.detect_sentiment(Text=text, LanguageCode='en')
            r = np.zeros(3)
            r[0] = ret['SentimentScore']['Negative']
            r[1] = ret['SentimentScore']['Neutral'] + ret['SentimentScore']['Mixed']
            r[2] = ret['SentimentScore']['Positive']
            scores.append(r)
        return np.array(scores)


    def predict_proba_azure3(self, exs):
        print('Azure: predicting %d examples' % len(exs))
        headers = {
        # Request headers
        'Content-Type': 'application/json',
        #         'Ocp-Apim-Subscription-Key': '3045416f7b4e4e73a58209761510341b',
        # 'Ocp-Apim-Subscription-Key': '1ddbdc586002484f8d4ad20c1ef092c5',
            # 'Ocp-Apim-Subscription-Key': 'c7c3c9d329aa49639d5706d11b313721'
            'Ocp-Apim-Subscription-Key': '94990b170d4c47c29c8a23facf39a22d'
        }
        params = urllib.parse.urlencode({
            # Request parameters
            'showStats': 'false',
        })
        body = json.dumps({'documents': [{'id': str(i), 'text': x, 'language': 'en'} for i, x in enumerate(exs)]})
        try:
            conn = http.client.HTTPSConnection('westus2.api.cognitive.microsoft.com')
#             conn.request("POST", "/text/analytics/v2.1/sentiment?%s" % params, body, headers)
            conn.request("POST", "/text/analytics/v3.0-preview.1/sentiment?%s" % params, body, headers)
            response = conn.getresponse()
            azureresp = response.read()
            conn.close()
        except Exception as e:
            print("[Errno {0}] {1}".format(e.errno, e.strerror))
            print(exs)
        try:
            pps = np.array([[x['documentScores'][a] for a in ['negative', 'neutral', 'positive']] for x in json.loads(azureresp)['documents']])
        except:
            print(json.loads(azureresp))
            raise Exception()
        return pps
    def predict_proba_azure2(self, exs):
        print('Azure: predicting %d examples' % len(exs))
        headers = {
        # Request headers
        'Content-Type': 'application/json',
        #         'Ocp-Apim-Subscription-Key': '3045416f7b4e4e73a58209761510341b',
        # 'Ocp-Apim-Subscription-Key': '1ddbdc586002484f8d4ad20c1ef092c5',
            # 'Ocp-Apim-Subscription-Key': 'c7c3c9d329aa49639d5706d11b313721'
            'Ocp-Apim-Subscription-Key': '94990b170d4c47c29c8a23facf39a22d'
        }
        params = urllib.parse.urlencode({
            # Request parameters
            'showStats': 'false',
        })
        body = json.dumps({'documents': [{'id': str(i), 'text': x, 'language': 'en'} for i, x in enumerate(exs)]})
        try:
            conn = http.client.HTTPSConnection('westus2.api.cognitive.microsoft.com')
            conn.request("POST", "/text/analytics/v2.1/sentiment?%s" % params, body, headers)
            # conn.request("POST", "/text/analytics/v3.0-preview.1/sentiment?%s" % params, body, headers)
            response = conn.getresponse()
            azureresp = response.read()
            conn.close()
        except Exception as e:
            print("[Errno {0}] {1}".format(e.errno, e.strerror))
            print(exs)
        pps = [x['score'] for x in json.loads(azureresp)['documents']]
        return np.array(pps)
    def predict_proba_google(self, exs):
        type_ = language_v1.Document.Type.PLAIN_TEXT
        language = "en"
        encoding_type = language_v1.EncodingType.UTF8
        scores = []
        for text in exs:
            document = {"content": text, "type_": type_, "language": language}
            annotations = self.gclient.analyze_sentiment(request = {'document': document, 'encoding_type': encoding_type})
            score = 0.5 + annotations.document_sentiment.score / 2.
            scores.append(score)
        return np.array(scores)




    # def predict_proba_google(self, exs):
    #     print('Google: predicting %d examples' % len(exs))
    #     scores = []
    #     for text in exs:
    #         document = types.Document(
    #         content=text,
    #         language='en',
    #         type=enums.Document.Type.PLAIN_TEXT)
    #         annotations = self.gclient.analyze_sentiment(document=document)
    #         score = 0.5 + annotations.document_sentiment.score / 2.
    #         scores.append(score)
    #     return np.array(scores)



import sys
from tqdm.auto import tqdm
import os
import pickle
import urllib
import json
import http
import aiohttp
import asyncio
import numpy as np
import time
import requests


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

class SentimentModel:
    def __init__(self, pred_file, key, batch_size=10, wait_time=1):
        self.preds = {}
        self.post_fn = None
        self.key = key
        if os.path.exists(pred_file):
            self.preds = pickle.load(open(pred_file, 'rb'))
        self.pred_file = pred_file
        self.pp_fn = self.async_predict_proba_azure
        self.batch_size = batch_size
        self.wait_time = wait_time

    def binary_to_three(self, pr):
        # This is what google recommends
        margin_neutral = 0.25
        mn = margin_neutral / 2.
        pp = np.zeros((pr.shape[0], 3))
        neg = pr < 0.5 - mn
        pp[neg, 0] = 1 - pr[neg]
        pp[neg, 2] = pr[neg]
        pos = pr > 0.5 + mn
        pp[pos, 0] = 1 - pr[pos]
        pp[pos, 2] = pr[pos]
        neutral_pos = (pr >= 0.5) * (pr < 0.5 + mn)
        pp[neutral_pos, 1] = 1 - (1 / margin_neutral) * np.abs(pr[neutral_pos] - 0.5)
        pp[neutral_pos, 2] = 1 - pp[neutral_pos, 1]
        neutral_neg = (pr < 0.5) * (pr > 0.5 - mn)
        pp[neutral_neg, 1] = 1 - (1 / margin_neutral) * np.abs(pr[neutral_neg] - 0.5)
        pp[neutral_neg, 0] = 1 - pp[neutral_neg, 1]
        return (pp.T / pp.sum(1)).T # normalize to sum to 1

    def predict_proba(self, exs, silent=False):
        loop = asyncio.get_event_loop()
        to_pred = [x for x in exs if x not in self.preds]
        chunked = list(chunks(to_pred, self.batch_size))
        pps_chunked = loop.run_until_complete(asyncio.gather(*[self.pp_fn(docs) for docs in chunked]))
        for docs,pps in zip(chunked,pps_chunked):
            # pps = loop.run_until_complete(self.pp_fn(docs))
            for x, pp in zip(docs, pps):
                self.preds[x] = pp
            time.sleep(self.wait_time)
        ret = np.array([self.preds.get(x) for x in exs])
        pickle.dump(self.preds, open(self.pred_file, 'wb'))
        if self.post_fn:
            ret = self.post_fn(ret)
        return (ret.T / ret.sum(1)).T # normalize to sum to 1

    def predict_proba_sync(self, exs, silent=False):
        to_pred = [x for x in exs if x not in self.preds]
        chunked = list(chunks(to_pred, self.batch_size))
        for docs in tqdm(chunked, disable=silent):
            pps = self.predict_proba_azure(docs)
            for x, pp in zip(docs, pps):
                self.preds[x] = pp
            time.sleep(self.wait_time)
        ret = np.array([self.preds.get(x) for x in exs])
        pickle.dump(self.preds, open(self.pred_file, 'wb'))
        if self.post_fn:
            ret = self.post_fn(ret)
        return ret

    def predict_and_confidences(self, exs):
        confs = self.predict_proba(exs)
        preds = np.argmax(confs, axis=1)
        return preds, confs

    async def async_predict_proba_azure(self, exs):
        print('Azure: predicting %d examples' % len(exs))
        headers = {
        # Request headers
        'Content-Type': 'application/json',
            'Ocp-Apim-Subscription-Key': self.key,
        }
        params = urllib.parse.urlencode({
            # Request parameters
            'showStats': 'false',
        })
        body = json.dumps({'documents': [{'id': str(i), 'text': x, 'language': 'en'} for i, x in enumerate(exs)]})

        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post('https://westus2.api.cognitive.microsoft.com/text/analytics/v3.1/sentiment?opinionMining=false%s" % params', data=body) as resp:
                azureresp = await resp.text()
        # print(azureresp)
        try:
            pps = np.array([[x['confidenceScores'][a] for a in ['negative', 'neutral', 'positive']] for x in json.loads(azureresp)['documents']])
        except:
            print(json.loads(azureresp))
            raise Exception()
        return pps

    def predict_proba_azure(self, exs):

        print('Azure: predicting %d examples' % len(exs))
        headers = {
        # Request headers
        'Content-Type': 'application/json',
            'Ocp-Apim-Subscription-Key': self.key,
        }
        params = urllib.parse.urlencode({
            # Request parameters
            'showStats': 'false',
        })
        body = json.dumps({'documents': [{'id': str(i), 'text': x, 'language': 'en'} for i, x in enumerate(exs)]})
        try:
            conn = http.client.HTTPSConnection('westus2.api.cognitive.microsoft.com')
#             conn.request("POST", "/text/analytics/v2.1/sentiment?%s" % params, body, headers)
            conn.request("POST", "/text/analytics/v3.1-preview.3/sentiment?opinionMining=false%s" % params, body, headers)
            response = conn.getresponse()
            azureresp = response.read()
            conn.close()
        except Exception as e:
            print("[Errno {0}] {1}".format(e.errno, e.strerror))
            print(exs)
        try:
            pps = np.array([[x['confidenceScores'][a] for a in ['negative', 'neutral', 'positive']] for x in json.loads(azureresp)['documents']])
        except:
            print(json.loads(azureresp))
            raise Exception()
        return pps

    def predict_proba_azurea(self, exs):
        print('Azure: predicting %d examples' % len(exs))
        headers = {
        # Request headers
        'Content-Type': 'application/json',
            'Ocp-Apim-Subscription-Key': self.key,
        }
        params = urllib.parse.urlencode({
            # Request parameters
            'showStats': 'false',
        })
        body = json.dumps({'documents': [{'id': str(i), 'text': x, 'language': 'en'} for i, x in enumerate(exs)]})
        try:
            conn = http.client.HTTPSConnection('westus2.api.cognitive.microsoft.com')
#             conn.request("POST", "/text/analytics/v2.1/sentiment?%s" % params, body, headers)
            conn.request("POST", "/text/analytics/v3.1-preview.3/sentiment?opinionMining=false%s" % params, body, headers)
            response = conn.getresponse()
            azureresp = response.read()
            conn.close()
        except Exception as e:
            print("[Errno {0}] {1}".format(e.errno, e.strerror))
            print(exs)
        if not json.loads(azureresp)['documents']:
            print(json.loads(azureresp))
            raise Exception()
        try:
            pps = np.array([[x['confidenceScores'][a] for a in ['negative', 'neutral', 'positive']] for x in json.loads(azureresp)['documents']])
        except:
            print(json.loads(azureresp))
            raise Exception()
        return pps

class Translator:
    def __init__(self, pred_file, key, from_language='en', to_language='pt-br', batch_size=1000, wait_time=0):
        self.preds = {}
        self.post_fn = None
        self.key = key
        if os.path.exists(pred_file):
            self.preds = pickle.load(open(pred_file, 'rb'))
        self.pred_file = pred_file
        self.pred_fn = self.translate
        self.batch_size = batch_size
        self.wait_time = wait_time
        self.from_language = from_language
        self.to_language = to_language

    def predict_proba(self, exs):
        to_pred = [x for x in exs if x not in self.preds]
        chunked = list(chunks(to_pred, self.batch_size))
        for docs in tqdm(chunked):
            pps = self.pred_fn(docs)
            for x, translation in zip(docs, pps):
                self.preds[x] = translation
            pickle.dump(self.preds, open(self.pred_file, 'wb'))
            if self.wait_time:
                time.sleep(self.wait_time)
        ret = np.array([self.preds.get(x) for x in exs])
        if self.post_fn:
            ret = self.post_fn(ret)
        return ret

    def translate(self, exs):
        print('Azure: predicting %d examples' % len(exs))
        subscription_key = self.key
        headers = {
            'Ocp-Apim-Subscription-Key': subscription_key,
            'Ocp-Apim-Subscription-Region': 'westus2',
            'Content-type': 'application/json',
            # 'X-ClientTraceId': str(uuid.uuid4())
        }
        params = {
            'api-version': '3.0',
            'from': self.from_language,
            'to': [self.to_language]
        }
        body = [{'text': x} for x in exs]
        endpoint = 'https://api.cognitive.microsofttranslator.com/translate'
        try:
            request = requests.post(endpoint, params=params, headers=headers, json=body)
            response = request.json()
            return [x['translations'][0]['text'] for x in response]
        except Exception as e:
            print('ERROR!')
            print(e)
            # print("[Errno {0}] {1}".format(e.errno, e.strerror))
            print(exs[:10])
            print(response)

class CogNER:
    def __init__(self, pred_file, key, batch_size=1000, wait_time=0):
        self.preds = {}
        self.key = key
        if os.path.exists(pred_file):
            self.preds = pickle.load(open(pred_file, 'rb'))
        self.pred_file = pred_file
        self.pred_fn = self.ner
        self.batch_size = batch_size
        self.wait_time = wait_time
        self.post_fn = None

    def predict_proba(self, exs):
        to_pred = [x for x in exs if x not in self.preds]
        chunked = list(chunks(to_pred, self.batch_size))
        for docs in tqdm(chunked):
            pps = self.pred_fn(docs)
            for x, translation in zip(docs, pps):
                self.preds[x] = translation
            pickle.dump(self.preds, open(self.pred_file, 'wb'))
            if self.wait_time:
                time.sleep(self.wait_time)
        ret = np.array([self.preds.get(x) for x in exs])
        if self.post_fn:
            ret = self.post_fn(ret)
        return ret

    def ner(self, exs):
        print('Azure: predicting %d examples' % len(exs))
        subscription_key = self.key
        headers = {
            'Ocp-Apim-Subscription-Key': subscription_key,
            'Ocp-Apim-Subscription-Region': 'westus2',
            'Content-type': 'application/json',
            # 'X-ClientTraceId': str(uuid.uuid4())
        }
        params = {}
        body = json.dumps({'documents': [{'id': str(i), 'text': x, 'language': 'en'} for i, x in enumerate(exs)]})
        conn = http.client.HTTPSConnection('westus2.api.cognitive.microsoft.com')
        try:
            conn.request("POST", "/text/analytics/v3.2-preview.1/entities/recognition/general?%s" % params, body, headers)
            response = conn.getresponse()
            azureresp = response.read()
            response = json.loads(azureresp)
            conn.close()
            return [x['entities'] for x in json.loads(azureresp)['documents']]
        except Exception as e:
            print('ERROR!')
            print(e)
            # print("[Errno {0}] {1}".format(e.errno, e.strerror))
            print(exs[:10])
            print(response)



import pandas as pd
import uuid
import logging
import random
import os
import io
import time
import numpy as np
import pandas as pd
from ._prompt_builder import PromptBuilder
from ._test_tree_browser import TestTreeBrowser, is_subtopic
from ._model import Model
from ._topic_model import TopicLabelingModel, TopicMembershipModel
import adatest
from pathlib import Path
import torch

# class TestTreeIterator():
#     def __init__(self, test_tree):
#         self.test_tree = test_tree
#         self.position = 0

#     def __next__(self):
#         if self.position >= len(self.test_tree):
#             raise StopIteration
#         else:
#             self.position += 1
#             return self.test_tree.iloc[self.position - 1]

class TestTree():
    """ A hierarchically organized set of tests represented as a DataFrame.

    This represents a hierarchically organized set of tests that all target a specific class of models (such as sentiment
    analysis models, or translation models). To interact with a test tree you can use either the `__call__` method to
    view and create tests directly in a Jupyter notebook, or you can call the `serve` method to launch a standalone
    webserver. A TestTree object also conforms to most of the standard pandas DataFrame API.
    """

    def __init__(self, tests=None, index=None, compute_embeddings=False, ensure_topic_markers=True, **kwargs):
        """ Create a new test tree.

        Parameters
        ----------
        tests : str or DataFrame or list or tuple or None
            The tests to load as a test tree. If a string is provided, it is assumed to be a path to a CSV file containing
            the tests. If tests is a tuple of two elements, it is assumed to be a dataset of (data, labels) which will be used to build a test tree.
            Otherwise tests is passed to the pandas DataFrame constructor to load the tests as a DataFrame.

        index : list or list-like or None
            Assigns an index to underlying tests frame, or auto generates if not provided.

        compute_embeddings: boolean
            If True, use the global adatest.embed to build embeddings of tests in the TestTree.

        kwargs : dict
            Additional keyword arguments are passed to the pandas DataFrame constructor.
        """

        # the canonical ordered list of test tree columns
        column_names = ['topic', 'input', 'output', 'label', 'labeler', 'description']

        # create a new test tree in memory
        if tests is None:
            self._tests = pd.DataFrame([], columns=column_names, dtype=str)
            self._tests_location = None

        # create a new test tree on disk (lazily saved)
        elif isinstance(tests, str) and not os.path.isfile(tests):
            self._tests = pd.DataFrame([], columns=column_names)
            self._tests_location = tests

        # load the test tree from a file or IO stream
        elif isinstance(tests, str) or isinstance(tests, io.TextIOBase):
            self._tests_location = tests
            if os.path.isfile(tests) or isinstance(tests, io.TextIOBase):
                self._tests = pd.read_csv(tests, index_col=0, dtype=str, keep_default_na=False)
            else:
                raise Exception(f"The provided tests file is not supported: {tests}")

        elif isinstance(tests, tuple) and len(tests) == 2: # Dataset loader TODO: fix this for topic models
            # column_names = ['topic', 'type', 'value1', 'value2', 'value3', 'author', 'description', \
            # 'model value1 outputs', 'model value2 outputs', 'model value3 outputs', 'model score']

            self._tests = pd.DataFrame(columns=column_names)
            self._tests_location = None

            self._tests['input'] = tests[0]
            self._tests['output'] = tests[1]

            # Constants
            self._tests['topic'] = ''
            self._tests['label'] = ''
            self._tests['labeler'] = "dataset"
            self._tests['description'] = ''

        else:
            if index is None:
                index = [uuid.uuid4().hex for _ in range(len(tests))]
            self._tests = pd.DataFrame(tests, **kwargs)
            self._tests.index = index
            self._tests_location = None

        # # ensure auto saving is possible when requested
        # if auto_save and self._tests_location is None:
        #     raise Exception("auto_save=True is only supported when loading from a file or IO stream")
        # self.auto_save = auto_save

        # ensure we have required columns
        for c in ["input", "output", "label"]:
            if c not in self._tests.columns:
                raise Exception("The test tree being loaded must contain a '"+c+"' column!")

        # fill in any other missing columns
        for column in ["topic", "description"]:
            if column not in self._tests.columns:
                self._tests[column] = ["" for _ in range(self._tests.shape[0])]
        if "labeler" not in self._tests.columns:
            self._tests["labeler"] = ["imputed" for _ in range(self._tests.shape[0])]

        # ensure that all topics have a topic_marker entry
        if ensure_topic_markers:
            self.ensure_topic_markers()

        # drop any duplicate index values
        self._tests = self._tests.groupby(level=0).first()

        # drop any duplicate rows
        self._tests.drop_duplicates(["topic", "input", "output", "labeler"], inplace=True)

        # put the columns in a consistent order
        self._tests = self._tests[column_names + [c for c in self._tests.columns if c not in column_names]]

        if compute_embeddings:
            self._cache_embeddings()

        # replace any invalid topics with the empty string
        for i, row in self._tests.iterrows():
            if not isinstance(row.topic, str) or not row.topic.startswith("/"):
                self._tests.loc[i, "topic"] = ""

        self._topic_labeling_models = {}
        self._topic_membership_models = {}

        # # keep track of our original state
        # if self.auto_save:
        #     self._last_saved_tests = self._tests.copy()

    def ensure_topic_markers(self):
        marked_topics = {t: True for t in set(self._tests.loc[self._tests["label"] == "topic_marker"]["topic"])}
        for topic in set(self._tests["topic"]):
            parts = topic.split("/")
            for i in range(1, len(parts)+1):
                parent_topic = "/".join(parts[:i])
                if parent_topic not in marked_topics:
                    self._tests.loc[uuid.uuid4().hex] = {
                        "label": "topic_marker",
                        "topic": parent_topic,
                        "labeler": "imputed",
                        "input": "",
                        "output": "",
                        "description": ""
                    }
                    marked_topics[parent_topic] = True

    def __getitem__(self, key):
        """ TestSets act just like a DataFrame when sliced. """
        subset = self._tests[key]
        if hasattr(subset, 'columns') and len(set(["topic", "input", "output", "label"]) - set(subset.columns)) == 0:
            return self.__class__(subset, index=subset.index)
        return subset

    def __setitem__(self, key, value):
        """ TestSets act just like a DataFrame when sliced, including assignment. """
        self._tests[key] = value

    # all these methods directly expose the underlying DataFrame API
    @property
    def loc(self):
        return TestTreeLocIndexer(self)
    @property
    def iloc(self):
        return TestTreeILocIndexer(self)
    @property
    def index(self):
        return self._tests.index
    @property
    def columns(self):
        return self._tests.columns
    @property
    def shape(self):
        return self._tests.shape
    @property
    def iterrows(self):
        return self._tests.iterrows
    @property
    def groupby(self):
        return self._tests.groupby
    @property
    def drop(self):
        return self._tests.drop
    @property
    def insert(self):
        return self._tests.insert
    @property
    def copy(self):
        return self._tests.copy
    @property
    def sort_values(self):
        return self._tests.sort_values
    
    # NOTE: Can't delegate to df.append as it is depricated in favor of pd.concat, which we can't use due to type checks 
    def append(self, test_tree, axis=0):
        if isinstance(test_tree, pd.DataFrame):
            self._tests = pd.concat([self._tests, test_tree], axis=axis)
        elif isinstance(test_tree, TestTree):
            self._tests = pd.concat([self._tests, test_tree._tests], axis=axis)
        elif isinstance(test_tree, dict):
            # check if the values are strings or lists of strings
            if any([isinstance(v, str) for v in test_tree.values()]):
                self._tests = pd.concat([self._tests, pd.DataFrame({k: [test_tree[k]] for k in test_tree}, index=[uuid.uuid4().hex])], axis=axis)
            else:
                self._tests = pd.concat([self._tests, pd.DataFrame(test_tree)], axis=axis)

        #self.deduplicate()
        #self.compute_embeddings()
        return None # TODO: Rethink append logic -- return copy vs. in place update?

    def __len__(self):
        return self._tests.__len__()
    def __setitem__(self, key, value):
        return self._tests.__setitem__(key, value)
    def to_csv(self, file=None):
        no_suggestions = self._tests.loc[["/__suggestions__" not in topic for topic in self._tests["topic"]]]
        if file is None:
            no_suggestions.to_csv(self._tests_location)
        else:
            no_suggestions.to_csv(file)

    def topic(self, topic):
        """ Return a subset of the test tree containing only tests that match the given topic.

        Parameters
        ----------
        topic : str
            The topic to filter the test tree by.
        """
        ids = [id for id, test in self._tests.iterrows() if is_subtopic(topic, test.topic)]
        return self.loc[ids]

    def adapt(self, scorer=None, generator=adatest.generators.OpenAI(), auto_save=False, user="anonymous", recompute_scores=False, drop_inactive_score_columns=False,
              max_suggestions=100, suggestion_thread_budget=0.5, prompt_builder=PromptBuilder(), active_generator="default", starting_path="",
              score_filter=-1e10, topic_model_scale=0): # TODO: remove active_generator and replace with the ability to set the generator?
        """ Apply this test tree to a scorer/model and browse/edit the tests to adapt them to the target model.

        Applying a test tree to a target model (wrapped by a scorer) creates a TestTreeBrowser object that can be used to
        browse the tree and add new tests to adapt it to the target model.
        
        Parameters
        ----------
        scorer : adatest.Scorer or callable
            The scorer (that wraps a model) to used to score the tests. If a function is provided, it will be wrapped in a scorer.
            Passing a dictionary of scorers will score multiple models at the same time. Note that the models are expected to take
            a list of strings as input, and output either a classification probability vector or a string.

        generator : adatest.Generator or dict[adatest.Generators]
            A source to generate new tests from. Currently supported generator types are language models, existing test trees, or datasets.

        auto_save : bool
            Whether to automatically save the test tree after each edit.

        user : str
            The user name to author new tests with.

        recompute_scores : bool
            Whether to recompute the scores of the tests that already have score values in the test tree.

        drop_inactive_score_columns : bool
            Whether to drop the score columns in the test tree that do not match any of the passed scorers.

        max_suggestions : int
            The maximum number of suggestions to generate each time the user asks for test suggestions.

        suggestion_thread_budget : float
            This controls how many parallel suggestion processes to use when generating suggestions. A value of 0 means we create
            no parallel threads (i.e. we use a single thread), 0.5 means we create as many parallel threads as possible without
            increase the number of tokens we process by more than 50% (1.5 would mean 150%, etc.). Each thread process will use a
            different randomized LM prompt for test generation, so more threads will result in more diversity, but come at the cost
            of reading more prompt variations.

        prompt_builder : adatest.PromptBuilder
            A prompt builder to use when generating prompts for new tests. This object controls how the LM prompts
            are created when generating new tests.

        active_generator : "default", or a key name if generators is a dictionary
            Which generator from adatest.generators to use when generating new tests. This should always be set to "default" if
            generators is just a single generator and not a dictionary of generators.

        starting_path : str
            The path to start browsing the test tree from.
        """

        # build the test tree browser
        return TestTreeBrowser(
            self,
            scorer=scorer,
            generators=generator,
            auto_save=auto_save,
            user=user,
            recompute_scores=recompute_scores,
            drop_inactive_score_columns=drop_inactive_score_columns,
            max_suggestions=max_suggestions,
            suggestion_thread_budget=suggestion_thread_budget,
            prompt_builder=prompt_builder,
            active_generator=active_generator,
            starting_path=starting_path,
            score_filter=score_filter,
            topic_model_scale=topic_model_scale
        )

    def __repr__(self):
        return self._tests.__repr__()

    def _repr_html_(self):
        return self._tests._repr_html_()

    def deduplicate(self):
        """ Remove duplicate tests from the test tree.
        
        Note that we give precendence to the first test in a set of duplicates.
        """

        already_seen = {}
        drop_ids = []
        for id, test in self._tests.iterrows():
            k = test.topic + "|_ADA_JOIN_|" + test.input + "|_ADA_JOIN_|" + test.output
            if k in already_seen:
                drop_ids.append(id)
            else:
                already_seen[k] = True
        self._tests = self._tests.drop(drop_ids, axis=0)

    def _cache_embeddings(self):
        """ Pre-compute the embeddings for all test cases in the test tree.
        """

        # see what new embeddings we need to compute
        all_strings = []
        for id, test in self._tests.iterrows():
            if test.label == "topic_marker":
                parts = test.topic.rsplit("/", 1)
                str = parts[1] if len(parts) == 2 else ""
                all_strings.append(str)
            else:
                for str in [test.input, test.output]:
                    all_strings.append(str)

        # suggestions topics don't have topic markers so we check for them separately
        all_strings.append("__suggestions__")
        
        # we don't use the output of the embedding, just do this to get the embeddings cached
        adatest.embed(all_strings)

    def impute_labels(self):
        """ Impute missing labels in the test tree. """
        # TODO: this is just a random mock, it needs to implement real local topic models

        self._cache_embeddings()

        for id, test in self._tests.iterrows():
            if test.label == "":
                if self.topic_membership_model(test.topic)(test.input) == "off_topic":
                    self._tests.loc[id, "label"] = "off_topic"
                else:
                    self._tests.loc[id, "label"] = self.topic_labeling_model(test.topic)(test.input, test.output)
                self._tests.loc[id, "labeler"] = "imputed"

    # def predict_labels(self, topical_io_pairs):
    #     """ Return the label probabilities for a set of input-output pairs. [NOT USED RIGHT NOW]

    #     Parameters
    #     ----------
    #     io_pairs : list[(str, str)]
    #         A list of input-output pairs to score.

    #     Returns
    #     -------
    #     list[float]
    #         A list of label probabilities.
    #     """

    #     out = np.zeros(len(topical_io_pairs))

    #     to_embed = []
    #     topics = {}
    #     for i,(topic,input,output) in enumerate(topical_io_pairs):
    #         if topic not in topics:
    #             topics[topic] = []
    #         to_embed.append(input)
    #         to_embed.append(output)
    #         topics[topic].append((i, len(to_embed) - 2, len(to_embed) - 1))
    #     embeddings = adatest.embed(to_embed)
    #     features = [None for i in range(len(topical_io_pairs))]
    #     for topic in topics:
    #         features = []
    #         for i,ind1,ind2 in topics[topic]:
    #             features.append(np.hstack([embeddings[ind1], embeddings[ind2]]))
    #         features = np.vstack(features)

    #         label = np.array([v == "pass" for v in self.topic_model(topic)(features)], dtype=np.float32)
    #         for i, (j,_,_) in enumerate(topics[topic]):
    #             out[j] = label[i]

    #     return np.array(out)

    def topic_labeling_model(self, topic):
        topic = topic.replace("/__suggestions__", "") # predict suggestions using their parent topic label model
        if topic not in self._topic_labeling_models:
            self._topic_labeling_models[topic] = TopicLabelingModel(topic, self)
        return self._topic_labeling_models[topic]

    def topic_membership_model(self, topic):
        topic = topic.replace("/__suggestions__", "") # predict suggestions using their parent topic membership model
        if topic not in self._topic_membership_models:
            self._topic_membership_models[topic] = TopicMembershipModel(topic, self)
        return self._topic_membership_models[topic]

    def retrain_topic_labeling_model(self, topic):
        self._topic_labeling_models[topic] = TopicLabelingModel(topic, self)

    def retrain_topic_membership_model(self, topic):
        self._topic_membership_models[topic] = TopicMembershipModel(topic, self)

    def drop_topic(self, topic):
        """ Remove a topic from the test tree. """
        self._tests = self._tests.loc[self._tests["topic"] != topic]

class TestTreeLocIndexer():
    def __init__(self, test_tree):
        self.test_tree = test_tree

    def __repr__(self):
        return "TestTreeLocIndexer is an intermediate object for operating on TestTrees. Slice this object further to yield useful results."

    def __getitem__(self, key):
        # If all columns haven't changed, it's still a valid test tree
        # If columns have been dropped, return a Pandas object
        
        subset = self.test_tree._tests.loc[key]
        if hasattr(subset, 'columns') and len(set(["topic", "input", "output", "label"]) - set(subset.columns)) == 0:
            test_tree_slice = TestTree(subset, index=subset.index, ensure_topic_markers=False)
            test_tree_slice._tests_location = self.test_tree._tests_location
            return test_tree_slice
        else:
            return subset
    
    def __setitem__(self, key, value):
        self.test_tree._tests.loc[key] = value
    
class TestTreeILocIndexer():
    def __init__(self, test_tree):
        self.test_tree = test_tree

    def __repr__(self):
        return "TestTreeILocIndexer is an intermediate object for operating on TestTrees. Slice this object further to yield useful results."

    def __getitem__(self, key):
        # If all columns haven't changed, it's still a valid test tree
        # If columns have been dropped, return a Pandas object
        
        subset = self.test_tree._tests.iloc[key]
        if hasattr(subset, 'columns') and len(set(["topic", "input", "output", "label"]) - set(subset.columns)) == 0:
            test_tree_slice = TestTree(subset, ensure_topic_markers=False)
            test_tree_slice._tests_location = self.test_tree._tests_location
            return test_tree_slice
        else:
            return subset
    
    def __setitem__(self, key, value):
        self.test_tree._tests.iloc[key] = value

def _test_tree_from_dataset(X, y, model=None, time_budget=60, min_samples=100):
    column_names = ['topic', 'type' , 'value1', 'value2', 'value3', 'author', 'description', \
        'model value1 outputs', 'model value2 outputs', 'model value3 outputs', 'model score']

    test_frame = pd.DataFrame(columns=column_names)

    if model is None: # All we can do without a model defined at this stage.
        test_frame['value1'] = X
        test_frame['type'] = "{} should output {}"
        test_frame['value2'] = y

        # Constants
        test_frame['topic'] = ''
        test_frame['author'] = "dataset"
        test_frame['description'] = ''

        return TestTree(test_frame)
    
    if not isinstance(model, Model):
        model = Model(model)

    # Validate output types
    output_names = model.output_names   
    unknown_labels = set(y) - set(output_names)
    assert len(unknown_labels) == 0, f"Unknown labels found: {unknown_labels}. \
    Please update the label vector or output names property."

    # Time how long inference takes on a single sample
    try:
        start = time.time()
        _ = model(X[0:1])
        end = time.time()
    except Exception as e: # TODO: Improve this message
        raise ValueError(f"Training data cannot be evaluated by model. Error recieved: {e}.")

    # Ensure min_samples <= n_samples <= len(data) and computes in {time_budget} seconds
    n_samples = int(min(max(time_budget // (end - start), min_samples), len(X)))

    if n_samples < len(X):
        print(f"Only using {n_samples} samples to meet time budget of {time_budget} seconds.")
        # TODO: unify input types
        sample_indices = np.random.choice(np.arange(len(X)), n_samples, replace=False)
        X = [X[sample] for sample in sample_indices]
        y = [y[sample] for sample in sample_indices]

    # Build intermediate convenience frame
    df = pd.DataFrame(columns=['sample', 'label', 'label_proba', \
                                        'pred', 'pred_proba', 'largest_error', 'largest_error_proba'])
    df['sample'] = X
    df['label'] = y

    # model's current prediction
    raw_model_output = model(X)
    pred_indices = np.argsort(raw_model_output, axis=1)
    
    df['pred_proba'] = raw_model_output[range(len(pred_indices)), pred_indices[:, -1]]
    df['pred'] = [output_names[i] for i in pred_indices[:, -1]]

    label_lookup = {output:index for index, output in enumerate(output_names)}
    label_indices = [label_lookup[label] for label in y]
    df['label_proba'] = raw_model_output[range(len(label_indices)), label_indices]
    
    correct_predictions = df['pred'] == df['label']
    mispredictions = ~correct_predictions
    
    # For mispredicted samples, the largest error is the current prediction.
    df.loc[mispredictions, 'largest_error'] = df.loc[mispredictions, 'pred']
    df.loc[mispredictions, 'largest_error_proba'] = df.loc[mispredictions, 'pred_proba']
    
    # For correct samples, we use the 2nd highest class as the largest error.
    largest_errors = pred_indices[correct_predictions][:, -2]
    df.loc[correct_predictions, 'largest_error'] = [output_names[i] for i in largest_errors]
    df.loc[correct_predictions, 'largest_error_proba'] = raw_model_output[range(len(largest_errors)), largest_errors]

    df.index = [uuid.uuid4().hex for _ in range(len(df))]

    # If we have a scorer, we prefer to format tests as {X} should not output {largest_error}
    test_frame['value1'] = df['sample']
    test_frame['type'] = "{} should not output {}"
    test_frame['value2'] = df['largest_error']

    # Constants
    test_frame['topic'] = ''
    test_frame['author'] = "dataset"
    test_frame['description'] = ''

    test_frame.index = df.index
    
    return TestTree(test_frame, index=test_frame.index)


import time
from IPython.display import display, HTML
import openai
import numpy as np
import copy
import math
import pandas as pd
import torch
import json
import re
import collections

from transformers import MODEL_FOR_NEXT_SENTENCE_PREDICTION_MAPPING
from adatest.generators import TestTreeSource

from .comm import JupyterComm
import uuid
import pathlib
import copy
import re
import logging
import os
import io
import statistics
import checklist
import checklist.editor
from threading import Timer
from ._scorer import expand_template, clean_template, ClassifierScorer, GeneratorScorer, Scorer
from ._prompt_builder import PromptBuilder
import builtins
import adatest # Need to import like this to prevent circular dependencies
import urllib.parse

# from https://gist.github.com/walkermatt/2871026
def throttle(interval):
    """ Decorator that will postpone a functions
        execution so it does not run more than once per
        interval of time.
    """
    def decorator(fn):
        def throttled(*args, **kwargs):
            if not hasattr(throttled, "t") or not throttled.t.is_alive():
                def call_it():
                    fn(*args, **kwargs)
                throttled.t = Timer(interval, call_it)
                throttled.t.start()
        return throttled
    return decorator

log = logging.getLogger(__name__)

# import sys
# sys.stderr = open('/tmp/err.txt', 'w')

def file_log(*args):
    """ Used for logging when we don't have a stdout.

    This is used for debugging when we are being called from the javascript client. When we are
    called from the client we don't have a stdout attached to any cell in the notebook.

    Note to also catch errors you could do:
    import sys
    sys.stderr = open('err.txt', 'w')
    """
    #print(*args)
    f = open("log.txt", "a")  # append mode
    f.write(" ".join([str(msg) for msg in args])+"\n")
    f.flush()
    f.close()

def is_subtopic(topic, candidate):
    # Returns true if candidate is a subtopic of topic
    return True if re.search(r'^%s(/|$)' % topic.replace('+', r'\+'), candidate) else False

def matches_filter(test, filter_text: str):
    if filter_text is None or filter_text == "":
        return True
    else:
        return filter_text in test["input"] or filter_text in test["output"]


special_outputs = [
    "{MAX}"
]

valid_comparators = [
    "should not be",
    "should be",
    "should be the same as for"
]
FILLIN_PREFIX = '/Fill-ins'


# model("this is english") => []
# output_sampling="topk(10)"
# output_sampling="topp(10)"
# output_sampling="max"
# output_sampling="temperature(0.9)"

class TestTreeBrowser():
    """ Used for browsing and expanding a test tree.
    """

    def __init__(self, test_tree, scorer, generators, user, auto_save, recompute_scores, drop_inactive_score_columns,
                 max_suggestions, suggestion_thread_budget, prompt_builder, active_generator, starting_path,
                 score_filter, topic_model_scale):
        """ Initialize the TestTreeBrowser.
        
        See the __call__ method of TreeBrowser for parameter documentation.
        """

        self.test_tree = test_tree
        self.scorer = scorer
        self.generators = generators
        self.user = user
        self.auto_save = auto_save
        self.recompute_scores = recompute_scores
        self.drop_inactive_score_columns = drop_inactive_score_columns
        self.max_suggestions = max_suggestions
        self.suggestion_thread_budget = suggestion_thread_budget
        self.prompt_builder = prompt_builder
        self.active_generator = active_generator
        self.current_topic = starting_path
        self.score_filter = score_filter
        self.topic_model_scale = topic_model_scale
        self.filter_text = ""

        # convert single generator to the multi-generator format
        if not isinstance(self.generators, dict):
            self.generators = {'generator': self.generators}

        if adatest.default_generators is not None: # Merge default generators into generators
            self.generators = {**self.generators, **adatest.default_generators}

        # Find and cast any TestTrees in generators to TestTreeSource
        for generator_name, generator in self.generators.items():
            if isinstance(generator, adatest._test_tree.TestTree): # TODO: make this autoreload friendly
                self.generators[generator_name] = TestTreeSource(generator) 

        # get a reference to the active backend object
        if self.active_generator == "default":
            if isinstance(self.generators, dict):
                self._active_generator_obj = next(iter(self.generators.items()))[1]
            else:
                self._active_generator_obj = self.generators
        else:
            self._active_generator_obj = self.generators[self.active_generator]

        # if we are recomputing the scores then we erase all the old scores
        if recompute_scores is True:
            for c in self.test_tree.columns:
                if c.endswith("score"):
                    self.test_tree.drop(c, axis=1, inplace=True)

        # convert single scorer args to the multi-scorer format
        if callable(self.scorer):
            self.scorer = {"model": self.scorer}

        # note the score column of each scorer
        if isinstance(self.scorer, dict):
            self.score_columns = [k+" score" for k in self.scorer]
            for k in self.scorer:
                self.scorer[k] = Scorer(self.scorer[k])
        elif self.scorer is not None:
            self.score_columns = ["model score"]
            self.scorer = {"model": Scorer(self.scorer)}
        else:
            self.score_columns = []

        # find score columns that are not associated with a scorer
        for c in self.test_tree.columns:
            if c.endswith("score") and c not in self.score_columns:
                if drop_inactive_score_columns is True:
                    self.test_tree.drop(c, axis=1, inplace=True)
                else:
                    self.score_columns.append(c)

        # ensure that each scorer's score column is in the test tree dataframe
        for c in self.score_columns:
            if c not in self.test_tree.columns:
                self.test_tree[c] = ["" for _ in range(self.test_tree.shape[0])]

        # a unique identifier for this test set instance, used for UI connections
        self._id = uuid.uuid4().hex

        # these are all temporary state
        self._hidden_topics = {}
        self.comm = None

        # define our current mode, and set of supported modes
        self.mode = "tests" if self.test_tree.shape[0] > 0 else "topics"
        self.mode_options = [
            # "validity focused", # focus first on making valid in-topic tests, then secondarily on making those tests high scoring
            # "failure focused", # focus on making high scoring (failing) tests, then secondarily on making those tests valid and in-topic
            "tests", # suggest new tests
            "topics" # suggest new subtopics
        ]

        # apply all the scorers to the test tree (this updates the test tree)
        self._compute_embeddings_and_scores(self.test_tree, self.recompute_scores, overwrite_outputs=False, save_outputs=True)

        # # make sure all the tests have scores (if we have a scorer)
        # self._compute_embeddings_and_scores(self.test_tree)

        # ensure any test tree based generator has embeddings calculated
        if isinstance(self.generators, dict):
            for name, gen in self.generators.items():
                if getattr(gen, "gen_type", "") == "test_tree":
                    gen.source._cache_embeddings()

        # save the current state of the test tree
        self._auto_save()

        # init a blank set of suggetions
        # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
        self._suggestions_error = "" # tracks if we failed to generate suggestions

    def _repr_html_(self, prefix="", environment="jupyter", websocket_server=None):
        """ Returns the HTML interface for this browser.

        Parameters
        ----------
        prefix : str
            The URL prefix this test tree browser is being served from.

        environment : str
            The environment this test tree browser is being served from (jupyter or web).
        """

        # spin up a JupyterComm object if we are called directly (which we assume is in a notebook)
        if self.comm is None and environment == "jupyter":
            self.comm = JupyterComm(f'adatest_interface_target_{self._id}', self.interface_event)

        # dump the client javascript to the interface
        file_path = pathlib.Path(__file__).parent.absolute()
        with open(file_path / "resources" / "main.js", encoding="utf-8") as f:
            js_data = f.read()
        interface_html = f"""
<div id="adatest_container_{self._id}" style="width: 100%; all: initial;"></div>
<script type='text/javascript'>
  {js_data};
  AdaTestReactDOM.render(
    AdaTestReact.createElement(AdaTest, {{
      interfaceId: "{self._id}", environment: "{environment}", startingTopic: "{self.current_topic}", prefix: "{prefix}",
      websocket_server: {"undefined" if websocket_server is None else '"'+websocket_server+'"'},\
    }}, null),
    document.getElementById('adatest_container_{self._id}')
  );
</script>
"""
        return interface_html

    def display(self):
        """ Manually display the HTML interface.
        """
        display(HTML(self._repr_html_()))

    def interface_event(self, msg):
        """ Handle interface events from the client.

        Parameters
        ----------
        msg : dict
            The event messages from the client. Each key in the dictionary is a separate message to either the row
            specified by the key or to whole browser object if the key is 'browser'.
        """

        log.debug(f"interface_event({msg})")

        # loop over each event message
        for k in msg:
            if k == "browser":
                action = msg[k].get("action", None)
                
                # rewdraw the entire interface
                if action == "redraw":
                    self._refresh_interface()
                
                # generate a new set of suggested tests/topics
                elif action == "generate_suggestions":
                    self._clear_suggestions()
                    self.test_tree.retrain_topic_labeling_model(self.current_topic)
                    self.test_tree.retrain_topic_membership_model(self.current_topic)
                    self._generate_suggestions(filter=msg[k].get("filter", ""))
                    # if self._active_generator_obj is None:
                    #     self._suggestions_error = "No AdaTest generator has been set!"
                    # else:
                    #     self._generate_suggestions(filter=msg[k].get("filter", ""))
                        # # try:
                        # self.suggestions = self._generate_suggestions(filter=msg[k].get("filter", ""))
                        # # filter suggestions to relevant types
                        # if self.mode == "topics":
                        #     self.suggestions = self.suggestions[self.suggestions['type'] == "topic_marker"]
                        # elif self.mode == "tests":
                        #     self.suggestions = self.suggestions[self.suggestions['type'] != "topic_marker"]

                        # # Ensure valid suggestions exist.
                        # if self.suggestions.shape[0] > 0:  
                        #     self.suggestions.sort_values(self.score_columns[0], inplace=True, ascending=False, key=np.vectorize(score_max))
                        #     self._suggestions_error = ""
                        # else:
                        #     self._suggestions_error = True # Not sure if we should do this?
                        # except Exception as e:
                        #     log.debug(e)
                        #     self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
                        #     self._suggestions_error = True
                    self._refresh_interface()
                
                # change the current topic
                elif action == "change_topic":
                    self.current_topic = msg[k]["topic"]
                    # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)

                    # see if we have only topics are direct children, if so, we suggest topics, otherwise we suggest tests
                    has_direct_tests = False
                    has_known_subtopics = False
                    for k, test in self.test_tree.iterrows():
                        if test["topic"] == self.current_topic:
                            if test["label"] != "topic_marker":
                                has_direct_tests = True
                        elif is_subtopic(self.current_topic, test["topic"]):
                            has_known_subtopics = True
                    if not has_direct_tests and has_known_subtopics:
                        self.mode = "topics"
                    else:
                        self.mode = "tests"

                    self._refresh_interface()
                
                # clear the current set of suggestions
                elif action == "clear_suggestions":
                    self._clear_suggestions()
                    # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
                    self._refresh_interface()

                # add a new empty subtopic to the current topic
                elif action == "add_new_topic":
                    self.test_tree.loc[uuid.uuid4().hex] = {
                        "topic": self.current_topic + "/New topic",
                        "label": "topic_marker",
                        "input": "",
                        "output": "",
                        "labeler": self.user,
                        "description": ""
                    }
                    self._compute_embeddings_and_scores(self.test_tree)
                    self._auto_save()
                    self._refresh_interface()
                
                # add a new empty test to the current topic
                elif action == "add_new_test":

                    # add the new test row
                    row = {
                        "topic": self.current_topic,
                        "input": "New test", # The special value "New test" causes the interface to auto-select the text
                        "output": "",
                        "label": "",
                        "labeler": "imputed",
                        "description": ""
                    }
                    for c in self.score_columns:
                        row[c] = "__TOEVAL__"
                        row[c[:-6] + " raw outputs"] = "{}"
                    self.test_tree.loc[uuid.uuid4().hex] = row

                    self._compute_embeddings_and_scores(self.test_tree)
                    self._auto_save()
                    self._refresh_interface()

                # change which scorer/model is used for sorting tests
                elif action == "set_first_model":
                    name = msg[k]["model"]

                    # move to front of score columns
                    pos = len(self.test_tree.columns) - len(self.score_columns)
                    tmp = self.test_tree[name]
                    self.test_tree.drop(labels=[name], axis=1, inplace=True)
                    self.test_tree.insert(pos, name, tmp)

                    # update score columns list
                    self.score_columns.remove(name)
                    self.score_columns.insert(0, name)

                    self._auto_save()
                    self._refresh_interface()

                # change which generator is active
                elif action is None and "active_generator" in msg[k]:
                    self.active_generator = msg[k]["active_generator"]
                    self._active_generator_obj = self.generators[self.active_generator]

                # change between topics and tests
                elif action is None and "mode" in msg[k]:
                    self.mode = msg[k]["mode"]

                elif action == 'change_description':
                    id = msg[k]['topic_marker_id']
                    if id not in self.test_tree.index:
                        self.test_tree.loc[id, 'topic'] = "" # only the root topic would be missing from the tree
                        self.test_tree.loc[id, 'input'] = ""
                        self.test_tree.loc[id, 'output'] = ""
                        self.test_tree.loc[id, 'label'] = "topic_marker"
                    self.test_tree.loc[id, 'description'] = msg[k]['description']
                    self._auto_save()

                elif action == 'change_filter':
                    print("change_filter")
                    self.filter_text = msg[k]['filter_text']
                    self._refresh_interface()


            # if we are just updating a single row in tests then we only recompute the scores
            elif "topic" not in msg[k]:
                sendback_data = {}
                
                # convert template expansions into a standard value update
                if msg[k].get("action", "") == "template_expand":
                    template_value = self.templatize(self.test_tree.loc[k, msg[k]["value"]])
                    msg[k] = {msg[k]["value"]: template_value}
                    sendback_data[msg[k]["value"]] = template_value

                # update the row and recompute scores
                for k2 in msg[k]:
                    self.test_tree.loc[k, k2] = msg[k][k2]
                if "input" in msg[k] or "output" in msg[k]:
                    self.test_tree.loc[k, self.score_columns] = "__TOEVAL__"
                    self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs="output" not in msg[k])
                elif "label" in msg[k]:
                    #self.test_tree.retrain_topic_model(self.current_topic)
                    pass # SML: we could recompute the scores here but then that would change the output of stochastic output models
                    # sign = -1 if msg[k]["label"] == "pass" else 1
                    # self.test_tree.loc[k, self.score_columns] = ""#abs(float(self.test_tree.loc[k, self.score_columns])) #* sign
                    # self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs=False)

                # send just the data that changed back to the frontend
                sendback_data["scores"] = {c: [[k, v] for v in ui_score_parts(self.test_tree.loc[k, c], self.test_tree.loc[k, "label"])] for c in self.score_columns}
                outputs = {c: [[k, json.loads(self.test_tree.loc[k].get(c[:-6] + " raw outputs", "{}"))]] for c in self.score_columns}
                sendback_data["raw_outputs"] = outputs
                if "output" not in msg[k]: # if the output was given to us the client is managing its current state so we shouldn't send it back
                    sendback_data["output"] = self.test_tree.loc[k, "output"]
                sendback_data["label"] = self.test_tree.loc[k, "label"]
                sendback_data["labeler"] = self.test_tree.loc[k, "labeler"]
                sendback_data.update(self.test_display_parts(self.test_tree.loc[k]))
                self.comm.send({k: sendback_data})
                
                self._auto_save()

            # if we are just changing the topic
            elif "topic" in msg[k] and len(msg[k]) == 1:
                
                # move a test that is in the test tree
                if k in self.test_tree.index:
                    if msg[k]["topic"] == "_DELETE_": # this means delete the test
                        self.test_tree.drop(k, inplace=True)
                    else:
                        self.test_tree.loc[k, "topic"] = msg[k]["topic"]
                        self.test_tree.loc[k, "author"] = self.user
                
                # move a whole topic around
                else:
                    for id, test in self.test_tree.iterrows():
                        if is_subtopic(k, test.topic):
                            if msg[k]["topic"] == "_DELETE_":
                                self.test_tree.drop(id, inplace=True)
                            else:
                                self.test_tree.loc[id, "topic"] = msg[k]["topic"] + test.topic[len(k):]

                # Recompute any missing embeddings to handle any changes
                self._compute_embeddings_and_scores(self.test_tree)
                self._refresh_interface()
                self._auto_save()

            else:
                log.debug(f"Unable to parse the interface message: {msg[k]}")

    def _refresh_interface(self):
        """ Send our entire current state to the frontend interface.
        """

        # get the children of the current topic
        data = {}

        def create_children(data, tests, topic):
            children = []
            
            # add tests and topics to the data lookup structure
            for k, test in tests.iterrows():
                if is_subtopic(topic, test.topic):
                    
                    # add a topic
                    if test.label == "topic_marker":
                        if is_subtopic(topic, test.topic) and test.topic != topic:
                            name = test.topic[len(topic)+1:]
                            if "/" not in name: # only add direct children
                                data[test.topic] = {
                                    "label": test.label,
                                    "labeler": test.labeler,
                                    "description": "",
                                    "scores": {c: [] for c in self.score_columns},
                                    "topic_marker_id": k,
                                    "topic_name": name,
                                    "editing": test.topic.endswith("/New topic")
                                }
                                children.append(test.topic)
                    
                    # add a test
                    elif matches_filter(test, self.filter_text):
                        data[k] = {
                            "input": test.input,
                            "output": test.output,
                            "label": test.label,
                            "labeler": test.labeler,
                            "description": test.description,
                            "scores": {c: [[k, v] for v in ui_score_parts(test[c], test.label)] for c in self.score_columns},
                            "editing": test.input == "New test"
                        }

                        data[k]["raw_outputs"] = {c: [[k, safe_json_load(test.get(c[:-6] + " raw outputs", "{}"))]] for c in self.score_columns}
                        data[k].update(self.test_display_parts(test))
                        if test.topic == topic:
                            children.append(k)
            
            # fill in the scores for the child topics
            for k, test in tests.iterrows():
                if "/__suggestions__" not in test.topic and is_subtopic(topic, test.topic) and test.topic != topic:
                    child_topic = test.topic[len(topic):].split("/", 2)[1]
                    scores = data[topic+"/"+child_topic]["scores"]
                    for c in self.score_columns:
                        scores[c].extend([[k, v] for v in ui_score_parts(test[c], test.label)])

            # sort by score and always put new topics first
            def sort_key(id):
                try:
                    total = 0
                    count = 0
                    # offset = 0 if data[id]["label"] == "fail" else -1
                    for s in data[id]["scores"][self.score_columns[0]]:
                        val = score_max(s[1], nan_val=np.nan)
                        if not np.isnan(val) and val is not None:
                            total += val #+ offset
                            count += 1
                    if count == 0:
                        return 1e3
                    else:
                        return -total / count
                except Exception as e:
                    print(e)
                    print(id)
                    print(val)
            sorted_children = sorted(children, key=sort_key)
            sorted_children = sorted(sorted_children, key=lambda id: 0 if data[id].get("label", "") == "topic_marker" else 1) # put folders first
            sorted_children = sorted(sorted_children, key=lambda id: 1 if data[id].get("label", "") == "off_topic" else 0) # off topic last
            sorted_children = sorted(sorted_children, key=lambda id: 0 if id.endswith("/New topic") or data[id].get("value1", "") == "New test" else 1) # put new items first

            return sorted_children
        
        # get the children of the current topic
        children = create_children(data, self.test_tree, self.current_topic)
        suggestions_children = create_children(data, self.test_tree, self.current_topic + "/__suggestions__")

        # TODO: This is a complete hack to hide lower scoring suggestions when we are likely already in the exploit phase
        # this is just for users who don't know when to stop scrolling down...
        # SML: I expect we can delete this at some point?
        if self.score_filter == "auto":
            if len(children) < 10:
                score_filter = -1e12
            else:
                children_scores = sorted([np.max([score_max(x[1]) for x in data[key]['scores'][self.score_columns[0]]]) for key in children])
                suggestions_children_scores = sorted([np.max([score_max(x[1]) for x in data[key]['scores'][self.score_columns[0]]]) for key in suggestions_children])
                score_filter = children_scores[-5] - (children_scores[-1] - children_scores[-5]) * 0.2
                if len(suggestions_children_scores) > 0:
                    score_filter = min(score_filter, np.nanmax(suggestions_children_scores) - 1e-2)
        else:
            score_filter = self.score_filter

        # if self.scorer is not None:
        #     test_types = self.scorer[self.score_columns[0][:-6]].supported_test_types
        #     test_type_parts = {t: split_test_type(t) for t in self.scorer[self.score_columns[0][:-6]].supported_test_types}
        # else:
        #     test_types = []
        #     test_type_parts = {}

        topic_marker_id = self._get_topic_marker_id(self.current_topic)
        # compile the global browser state for the frontend
        data["browser"] = {
            "suggestions": suggestions_children,
            "tests": children,
            "user": self.user,
            "topic": self.current_topic,
            "topic_description": self.test_tree.loc[topic_marker_id]["description"] if topic_marker_id is not None else "",
            "topic_marker_id": topic_marker_id if topic_marker_id is not None else uuid.uuid4().hex,
            "score_filter": score_filter,
            "disable_suggestions": False,
            "read_only": False,
            "score_columns": self.score_columns,
            "suggestions_error": self._suggestions_error,
            "generator_options": [str(x) for x in self.generators.keys()] if isinstance(self.generators, dict) else [self.active_generator],
            "active_generator": self.active_generator,
            "mode": self.mode,
            "mode_options": self.mode_options,
            # "test_types": test_types,
            # "test_type_parts": test_type_parts,
        }

        self.comm.send(data)

    def _clear_suggestions(self):
        """ Clear the suggestions for the current topic.
        """
        ids = list(self.test_tree.index)
        for k in ids:
            if self.test_tree.loc[k, "topic"].startswith(self.current_topic + "/__suggestions__"):
                self.test_tree.drop(k, inplace=True)

    def generate_suggestions(self, topic=None, filter=""):
        if topic is not None:
            self.current_topic = topic
        self._clear_suggestions()
        self.test_tree.retrain_topic_labeling_model(self.current_topic)
        self.test_tree.retrain_topic_membership_model(self.current_topic)
        self._generate_suggestions(filter=filter)

    def _generate_suggestions(self, filter):
        """ Generate suggestions for the current topic.

        Parameters
        ----------
        filter : str
            The filter to apply to the tests while generating suggestions.
        """

        #--Backend-driven suggestions--

        # save a lookup we can use to detect duplicate tests
        test_map = {}
        for _, test in self.test_tree.iterrows():
            if test.label == "topic_marker":
                test_map[test.topic + " __topic_marker__"] = True
            else:
                test_map[test.topic + " __JOIN__ " + test.input] = True

        
        # validity focused (focus first on making valid in-topic tests, then secondarily on making those tests high scoring)
        # failure focused (focus on making high scoring (failing) tests, then secondarily on making those tests valid and in-topic)
        # topics (suggest new sub-topics)
        # file_name dataset (suggest tests based on samples from the provided dataset)


        # compute the maximum number of suggestion threads we can use given our suggestion_thread_budget
        p = self.prompt_builder.prompt_size
        budget = 1 + self.suggestion_thread_budget
        suggestion_threads = max(1, int(np.floor(budget * (p/(p+1) + 1/(p+1) * self.max_suggestions) - 1/(p+1) * self.max_suggestions) / (p/(p+1))))
        
        # generate the prompts for the backend
        prompts = self.prompt_builder(
            test_tree=self.test_tree,
            topic=self.current_topic,
            score_column=self.score_columns[0],
            repetitions=suggestion_threads,
            filter=filter,
            suggest_topics=self.mode == "topics"
        )

        # get the current topic description
        curr_topic_mask = (self.test_tree["topic"] == self.current_topic) & (self.test_tree["label"] == "topic_marker")
        if curr_topic_mask.sum() == 0:
            desc = ""
        else:
            desc = self.test_tree.loc[(self.test_tree["topic"] == self.current_topic) & (self.test_tree["label"] == "topic_marker")]["description"][0]

        # generate the suggestions
        generators = [self._active_generator_obj] + list(self.generators.values())
        for generator in generators:
            try:
                proposals = generator(prompts, self.current_topic, desc, self.mode, self.scorer, num_samples=self.max_suggestions // len(prompts) if len(prompts) > 0 else self.max_suggestions)
                break
            except ValueError:
                pass # try the next generator
        
        # all topics should be URI encoded
        if self.mode == "topics":
            proposals = [urllib.parse.quote(x) for x in proposals]
        
        # Build up suggestions catalog, unless generating from a test tree source.
        # NOTE: Doing safe checks for TestTree type in order to prevent circular imports
        if isinstance(proposals, pd.DataFrame) or proposals.__class__.__name__ == "TestTree":
            suggestions = proposals
            suggestions['topic'] = self.current_topic + "/__suggestions__" + suggestions['topic'].apply(lambda x: x[len(self.current_topic):] if x != "" else "")
            self.test_tree.append(suggestions)
            print("appended suggestions into self.test_tree")
            # assert False, "This needs to be fixed to dump into /__suggestions__"
        else:
            # suggestions = []
            test_map_tmp = copy.copy(test_map)
            for input in proposals:
                if self.mode == "topics" and ("/" in input or "\n" in input):
                    input = input.replace("/", " or ").replace("\n", " ") # topics can't have newlines or slashes in their names
                    input = input.replace("  ", " ").strip() # kill any double spaces we may have introduced
                    str_val = self.current_topic + "/" + input + " __topic_marker__"
                else:
                    str_val = self.current_topic + " __JOIN__ " + input
                if str_val not in test_map_tmp:
                    id = uuid.uuid4().hex
                    self.test_tree.loc[id, "topic"] = self.current_topic + "/__suggestions__" + ("/"+input if self.mode == "topics" else "")
                    self.test_tree.loc[id, "input"] = "" if self.mode == "topics" else input
                    self.test_tree.loc[id, "output"] = "__TOOVERWRITE__"
                    self.test_tree.loc[id, "label"] = "topic_marker" if self.mode == "topics" else ""
                    self.test_tree.loc[id, "labeler"] = "imputed"
                    self.test_tree.loc[id, "description"] = ""
                    for c in self.score_columns:
                        self.test_tree.loc[id, c] = "__TOEVAL__"

                    # s = {
                    #     "topic": self.current_topic + "/__suggestions__" + ("/"+input if self.mode == "topics" else ""),
                    #     "input": "" if self.mode == "topics" else input,
                    #     "output": "",
                    #     "label": "",
                    #     "labeler": "imputed",
                    #     "description": ""
                    # }
                    # for c in self.score_columns:
                    #     s[c] = ""
                    # suggestions.append(s)
                    if str_val is not None:
                        test_map_tmp[str_val] = True

            # suggestions = pd.DataFrame(suggestions, index=[uuid.uuid4().hex for _ in range(len(suggestions))], columns=self.test_tree.columns)
            self._compute_embeddings_and_scores(self.test_tree)

        # Filter invalid suggestions
        # if self.mode != "topics":
        #     suggestions = suggestions.dropna(subset=[self.score_columns[0]])

        # When we have outputs filled in by the scorer we might have more duplicates we need to remove
        # duplicates = []
        # for k,row in suggestions.iterrows():
        #     # str_val = row.topic + " " + test_type + " " + row.value1 + " " +  row.value2 + " " +  row.value3
        #     str_val = " ".join(builtins.filter(None, (row.topic, test_type, row.value1, row.value2, row.value3))) # Safely handles None
        #     if str_val in test_map:
        #         duplicates.append(k)
        #     test_map[str_val] = True
        # suggestions = suggestions.drop(duplicates)

        # if self.topic_model_scale != 0:
        #     self._add_topic_model_score(suggestions, topic_model_scale=self.topic_model_scale)
        # return suggestions

    def _get_topic_marker_id(self, topic):
        """
        Returns the id of the topic marker row for the given topic.
        Returns None if not found.
        """
        topic_marker_index_df = self.test_tree.index[(self.test_tree['topic'] == topic) & (self.test_tree['label'] == 'topic_marker')]
        topic_marker_index = topic_marker_index_df.tolist()[0] if len(topic_marker_index_df) > 0 else None
        return topic_marker_index

    def _add_topic_model_score(self, df, topic_model_scale):
        documents = []
        for k,s in df.iterrows():
            max_output = -10e8
            max_output_name = None
            for k,v in json.loads(s["score value1 outputs"]).items():
                if v > max_output:
                    max_output = v
                    max_output_name = k
            documents.append(f'"{s["value1"]}" > "{max_output_name}"')

        query = self._make_prompt(
            self.current_topic,
            prompt_size=20,
            include_value2=True
        )["prompt"]

        r = openai.Engine("davinci-instruct-beta").search(
            documents=documents,
            query=query
        )

        sim_scores = np.array([v["score"] for v in r["data"]])
        sim_scores -= np.mean(sim_scores)
        sim_scores /= np.std(sim_scores)

        for i, (k, row) in enumerate(df.iterrows()):
            row["score"] = float(row["score"]) + topic_model_scale * sim_scores[i]

    def _compute_embeddings_and_scores(self, tests, recompute=False, overwrite_outputs=False, save_outputs=False): # TODO: Rename/refactor/merge with _compute_scores?
        log.debug(f"compute_embeddings_and_scores(tests=<DataFrame shape={tests.shape}>, recompute={recompute})")

        for k in self.scorer:
            # determine which rows we need to evaluate
            eval_ids = []
            for i, (id, test) in enumerate(tests.iterrows()):
                if (recompute or test[k+" score"] == "__TOEVAL__") and test.label != "topic_marker" and test.label != "off_topic":
                    eval_ids.append(id)

            if len(eval_ids) > 0:

                # run the scorer
                new_outputs,scores = self.scorer[k](tests, eval_ids)

                # update the scores in the test tree
                current_outputs = tests["output"]
                for i,id in enumerate(eval_ids):
                    # tests.loc[id, k+" score"] = scores[i]

                    if not overwrite_outputs and current_outputs.loc[id] != "__TOOVERWRITE__" and current_outputs.loc[id] != new_outputs[i]:

                        # mark the current row as nan score (meaning the output does not match)
                        tests.loc[id, k+" score"] = np.nan

                        # add a new test where the model output does match if we are saving outputs
                        if save_outputs:
                            id_new = uuid.uuid4().hex
                            tests.loc[id_new, "topic"] = tests.loc[id, "topic"]
                            tests.loc[id_new, "input"] = tests.loc[id, "input"]
                            tests.loc[id_new, "output"] = new_outputs[i]
                            tests.loc[id_new, "labeler"] = "imputed"
                            tests.loc[id_new, "label"] = ""
                            tests.loc[id_new, k+" score"] = scores[i]
                    else:
                        tests.loc[id, "output"] = new_outputs[i]
                        tests.loc[id, k+" score"] = scores[i]

        tests.deduplicate() # make sure any duplicates we may have introduced are removed

        # reimpute missing labels
        tests.impute_labels() # TODO: ensure this method caches the local models and only reimputes when needed for each topic

    def _compute_scores(self, tests, recompute):
        """ Use the scorer(s) to fill in scores in the passed TestTree.

        Parameters
        ----------
        tests : TestTree
            The TestTree to fill in missing scores for.

        recompute : bool
            If True, recompute all scores. If False, only recompute scores that are missing.
        """
        
        log.debug(f"_compute_scores(tests=<TestTree shape={tests.shape}>, recompute={recompute})")

        # see which rows need scores computed
        if recompute or len(self.score_columns) == 0:
            new_score_mask = np.ones(tests.shape[0], dtype=np.bool)
        else:
            new_score_mask = np.array(tests[self.score_columns[0]].isnull()) | np.array(tests[self.score_columns[0]] == "")
        new_score_mask = new_score_mask & np.array(tests["label"] != "topic_marker", dtype=np.bool)

        if new_score_mask.sum() > 0:
            scores = {}
            tests_to_score = tests.loc[new_score_mask, ["topic", "input", "output", "label"]]
            
            # call the scorers
            blank_outputs = [{} for _ in range(tests_to_score.shape[0])]
            for k in self.scorer:
                scorer_output = self.scorer[k](tests_to_score)
                scores[k+" score"] = ["|".join(str(vv) for vv in v) for v in scorer_output["scores"]]
                scores[k+" value1 outputs"] = scorer_output.get("value1_outputs", blank_outputs)
                scores[k+" value2 outputs"] = scorer_output.get("value2_outputs", blank_outputs)
                scores[k+" value3 outputs"] = scorer_output.get("value3_outputs", blank_outputs)

            # copy the scores into the TestTree
            for k in scores:
                for i, j in enumerate(np.where(new_score_mask)[0]):
                    tests.loc[tests.index[j], k] = json.dumps(scores[k][i]) if isinstance(scores[k][i], dict) else scores[k][i]

            # copy outputs that may have been generated by the scorers over to the passed test tree
            for k in tests.index[new_score_mask]:
                tests.loc[k, "value1"] = tests_to_score.loc[k, "value1"]
                tests.loc[k, "value2"] = tests_to_score.loc[k, "value2"]
                tests.loc[k, "value3"] = tests_to_score.loc[k, "value3"]

    # def _load_dataset(self, time_budget=30, min_samples=100):
    #     '''Evaluate model on dataset and capture useful information.'''
    #     # TODO: Generalize to more dataset formats
    #     if self.dataset is None:
    #         return None
        
    #     model = self.scorer['model'].model

    #     # Unpack dataset object
    #     X, y = self.dataset[0], self.dataset[1]
    #     output_names = self.scorer['model'].output_names
        
    #     unknown_labels = set(y) - set(output_names)
    #     assert len(unknown_labels) == 0, f"Unknown labels found: {unknown_labels}. \
    #     Please update the label vector or output names property."

    #     # Time how long inference takes on a single sample
    #     try:
    #         start = time.time()
    #         _ = model(X[0:1])
    #         end = time.time()
    #     except Exception as e: # TODO: Improve this message
    #         raise ValueError(f"Training data cannot be evaluated by model. Error recieved: {e}.")


    #     # Ensure min_samples <= n_samples <= len(data) and computes in {time_budget} seconds
    #     n_samples = int(min(max(time_budget // (end - start), min_samples), len(X)))

    #     if n_samples < len(X):
    #         print(f"Only using {n_samples} samples to meet time budget of {time_budget} seconds.")
    #         # TODO: unify input types
    #         sample_indices = np.random.choice(np.arange(len(X)), n_samples, replace=False)
    #         X = [X[sample] for sample in sample_indices]
    #         y = [y[sample] for sample in sample_indices]

    #     # Build output frame
    #     df = pd.DataFrame(columns=['sample', 'label', 'label_proba', \
    #                                         'pred', 'pred_proba', 'largest_error', 'largest_error_proba'])
    #     df['sample'] = X
    #     df['label'] = y

    #     # model's current prediction
    #     raw_model_output = model(X)
    #     pred_indices = np.argsort(raw_model_output, axis=1)
        
    #     df['pred_proba'] = raw_model_output[range(len(pred_indices)), pred_indices[:, -1]]
    #     df['pred'] = [output_names[i] for i in pred_indices[:, -1]]

    #     label_lookup = {output:index for index, output in enumerate(output_names)}
    #     label_indices = [label_lookup[label] for label in y]
    #     df['label_proba'] = raw_model_output[range(len(label_indices)), label_indices]
        

    #     correct_predictions = df['pred'] == df['label']
    #     mispredictions = ~correct_predictions
        
    #     # For mispredicted samples, the largest error is the current prediction.
    #     df.loc[mispredictions, 'largest_error'] = df.loc[mispredictions, 'pred']
    #     df.loc[mispredictions, 'largest_error_proba'] = df.loc[mispredictions, 'pred_proba']
        
    #     # For correct samples, we use the 2nd highest class as the largest error.
    #     largest_errors = pred_indices[correct_predictions][:, -2]
    #     df.loc[correct_predictions, 'largest_error'] = [output_names[i] for i in largest_errors]
    #     df.loc[correct_predictions, 'largest_error_proba'] = raw_model_output[range(len(largest_errors)), largest_errors]

    #     df.index = [uuid.uuid4().hex for _ in range(len(df))]
    #     return df

    # def _convert_dataset_to_tests(self, dataset_frame): # TODO: Consider removing from class?
    #     '''Converts a loaded dataset into test formats.'''
        
    #     column_names = ['topic', 'type' , 'value1', 'value2', 'value3', 'author', 'description', \
    #     'model value1 outputs', 'model value2 outputs', 'model value3 outputs', 'model score']

    #     test_frame = pd.DataFrame(columns=column_names)

    #     # All tests currently formatted as not predicting the largest error.
    #     test_frame['value1'] = dataset_frame['sample']
    #     test_frame['type'] = "{} should not output {}"
    #     test_frame['value2'] = dataset_frame['largest_error']

    #     # Constants
    #     test_frame['topic'] = ''
    #     test_frame['author'] = "dataset"
    #     test_frame['description'] = ''

    #     test_frame.index = dataset_frame.index
        
    #     return test_frame # TODO: Cast this as a formal TestTree instead of dataframe

    def test_display_parts(self, test):
        
        # # find which template instantiation has the highest score (and so should be displayed)
        # ui_score_parts = test[self.score_columns[0]].split("|")
        # if len(ui_score_parts) == 1:
        #     max_score_ind = 0
        # else:
        #     max_score_ind = np.argmax([float(v) for v in test[self.score_columns[0]].split("|")])
        
        # if test.type == "{} should not be completed to become {}":
        #     value1_outputs = safe_json_load(test.get(self.score_columns[0][:-6] + " value1 outputs", "{}")).get("string", [""])
        #     if not isinstance(value1_outputs, list):
        #         value1_outputs = [value1_outputs]
        #     if len(value1_outputs) == 1:
        #         value1_outputs_str = value1_outputs[0]
        #     else:
        #         value1_outputs_str = "{" + "|".join(value1_outputs) + "}"

        #     # hispanic / 95f08d610df74679afc1217f03b9c1b3.template[0] "this is a test of[]" should not be completed to become "toxic"
        #     # {}

        #     out = []
        #     for i,option in enumerate(expand_template(test.value1 + "|||" + value1_outputs_str + "|||" + test.value2, keep_braces=False)):
        #         value1_disp,d_text1b,value2_disp = option.split("|||")
        #         out.append({
        #             "d_text1a": '"',
        #             "d_value1": "{}",
        #             "value1_disp": value1_disp,
        #             "d_text1b": d_text1b + '"',
        #             "d_text2a": '"',
        #             "d_value2": "{}",
        #             "value2_disp": value2_disp,
        #             "d_text2b": '"',
        #             "d_text3a": '',
        #             "d_value3": "",
        #             "d_text3b": ''
        #         })
        # else: # this is the default two-value test format that only varies in the select value
        out = [{
            "d_text1a": '"',
            "d_value1": "{}",
            "d_text1b": '"',
            "d_text2a": '"',
            "d_value2": "{}",
            "d_text2b": '"',
            "d_text3a": '',
            "d_value3": "",
            "d_text3b": ''
        }]
        
        return {
            "display_parts": out
        }

    def templatize(self, s):
        prompt = """INPUT: "Where are regular people on Twitter"
    OUTPUT: "Where are {regular|normal|sane|typical} people on {Twitter|Facebook|Reddit|Instagram}"
    ###
    INPUT: "Anyone who says this food tastes horrible is out of their mind"
    OUTPUT: "{Anyone|Someone|He|She} who says this food tastes {horrible|terrible|rotten} is out of their mind"
    ###
    INPUT: "great"
    OUTPUT: "{great|excellent|wonderful|superb|delightful}"
    ###
    INPUT: "If you haven't come here before, you probably live under a rock"
    OUTPUT: "If you haven't come here {before|in the past|before now|yet}, you probably {live under a rock|never get out|are a hermit|are isolated}"
    ###
    INPUT: "Only crazy people would say they had a lousy time"
    OUTPUT: "Only {crazy people|insane people|people with no sense|those out of their mind} would say they had a {lousy|terrible|bad|miserable} time"
    ###
    INPUT: "If I didn't come here again, I would be very happy for the rest of my life"
    OUTPUT: "If I didn't come {here|hereabouts|around here} {again|once more|all over again}, I would be very {happy|glad|pleased|elated} for the rest of my life"
    ###
    INPUT: "I don't know what James was talking about when they said they loved the food."
    OUTPUT: "I don't know what {James|John|Robert|Steve|Bob} was talking about when they {said they|stated that|claimed that|mentioned that} they {loved|liked|adored|appreciated} the food."
    ###
    INPUT: "new_input_value"
    OUTPUT: \""""
        prompt = prompt.replace("new_input_value",  s)
        response = openai.Completion.create(
            engine=self.engine, prompt=prompt, max_tokens=300,
            temperature=0.7, n=4, stop="\""
        )

        lines = [choice["text"] for choice in response["choices"]]
        options = []
        for line in lines:
            line = clean_template(line)
            valid = False
            for option in expand_template(line):
                if option == s:
                    valid = True
                    break
            if valid:
                options.append((-len(line), line))
        options.sort()
        log.debug(f"options = {options}")
        return options[0][1]
    
    def _auto_save(self):
        """ Save the current state of the model if we are auto saving.
        """
        if self.auto_save:
            self.test_tree.to_csv()

def score_max(s, nan_val=-1e3):
    if s == "" or s is None:
        return nan_val
    elif isinstance(s, str):
        return np.max([convert_float(v) for v in s.split("|")])
    elif np.isnan(s):
        return nan_val
    else:
        return np.max(s)

def ui_score_parts(s, label):
    """ Split a score into its parts and encode the label into the sign.
    
    Note this encoding is just used for passing scores to the UI (scores are not signed in the TestTree).
    """
    offset = 0
    if label == "pass":
        sign = 1
        offset = -1 - 1e-6
    elif label == "fail":
        sign = 1
        offset = 1e-6 # just so we have a positive number to encode that this was a failure
    else:
        sign = np.nan
    if isinstance(s, str):
        return [np.clip(offset + convert_float(v)*sign, -1, 1) for v in s.split("|")]
    else:
        return [np.clip(offset + s*sign, -1, 1)]

def convert_float(s):
    if s == "":
        return np.nan
    try:
        f = float(s)
    except ValueError:
        f = np.nan
    return f

def safe_json_load(input):
    if isinstance(input, float): # catch NaN's
        return {}
    else:
        return json.loads(input)

def split_test_type(test_type):
    part_names = ["text1", "value1", "text2", "value2", "text3", "value3", "text4"]
    parts = re.split(r"(\{\}|\[\])", test_type)
    part_values = ["" for _ in range(7)]
    for i, part in enumerate(parts):
        part_values[i] = part
    return {name: value for name,value in zip(part_names, part_values)}



def safe_mode(l):
    """ This just silences the error from a double mode from python <= 3.7.
    """
    try:
        return statistics.mode(l)
    except:
        return l[0]


import sklearn
import numpy as np
from sklearn import multioutput
from sklearn import preprocessing
from sklearn.linear_model import RidgeClassifierCV
from sklearn.linear_model import LogisticRegression
from sklearn.svm import LinearSVC
import adatest
import re


def is_subtopic(topic, candidate):
    # Returns true if candidate is a subtopic of topic
    return True if re.search(r'^%s(/|$)' % topic.replace('+', r'\+'), candidate) else False

class ConstantModel():
    def __init__(self, label):
        self.label = label
    def predict(self, embeddings):
        if not hasattr(embeddings[0], "__len__"):
            return self.label
        else:
            return [self.label] * len(embeddings)

class OutputNearestNeighborLabelModel():
    def __init__(self, embeddings, labels):
        embeddings[:,:embeddings.shape[1]//2] = 0 # zero out the embedding for the input value so we only depend on the output
        self.model = sklearn.neighbors.KNeighborsClassifier(1)
        self.model.fit(embeddings, labels)
    def predict(self, embeddings):
        embeddings[:,:embeddings.shape[1]//2] = 0
        return self.model.predict(embeddings)

class TopicLabelingModel:
    def __init__(self, topic, test_tree):
        self.topic = topic
        self.test_tree = test_tree

        # compute the weighting of samples for each test for this topic (SML: this is disabled below right now)
        # TODO: make this code shared correctly with prompt builder
        parts = topic.split("/")
        topic_scaling = np.ones(test_tree.shape[0])
        for i in range(1, len(parts)):
            prefix = "/".join(parts[:i+1])
            topic_scaling *= 1 + 9 * np.array([v.startswith(prefix) for v in test_tree["topic"]])
        topic_scaling += 1 * np.array([v == topic for v in test_tree["topic"]])

        # collect the embeddings
        # TODO: like in prompt builder we should have a max of how many samples we work with
        null_embedding = np.hstack(adatest.embed(["", ""]))
        embeddings = []
        labels = []
        output_map = {}
        perfect_output_map = True
        for i, (id, test) in enumerate(test_tree.iterrows()):
            if test.labeler == "imputed" or test.label == "topic_marker" or test.label == "off_topic":
                topic_scaling[i] = 0
                embeddings.append(null_embedding)
                labels.append("pass")
            else:
                embeddings.append(np.hstack(adatest.embed([test.input, test.output])))
                labels.append(test.label)

        if len(embeddings) == 0 and len(labels) == 0: # Handle empty testframe case
            self.model = ConstantModel("Unknown")
        else:
            embeddings = np.vstack(embeddings)

            # normalize the weights
            topic_scaling /= topic_scaling.max()
            topic_scaling *= topic_scaling == 1 # TODO: this is a hack to turn off our dependenence on out-of-topic samples (unless we have zero in-topic samples)

            # check if the labels are perfectly determined by the output value
            output_map = {}
            perfect_output_map = True
            for i, (id, test) in enumerate(test_tree.iterrows()):
                if topic_scaling[i] > 0:
                    if output_map.get(test.output, test.label) != test.label:
                        perfect_output_map = False
                        break
                    output_map[test.output] = test.label
            
            # degenerate case with only one label type
            # if len(set([l for i,l in enumerate(labels) if topic_scaling[i] > 0])) == 1:
            #     self.model = ConstantModel(labels[0])
            # el
            # if perfect_output_map:
            #     self.model = OutputNearestNeighborLabelModel(embeddings, labels)
            

            # fit a linear SVM the outputs don't 
            if len(set([l for i,l in enumerate(labels) if topic_scaling[i] > 0])) > 1:
                # we are in a highly overparametrized situation, so we use a linear SVC to get "max-margin" based generalization
                # TODO: SML: It seems to me that the SVC seems to do very well as long as there are no "errors" in the data labels. But it will
                # do very poorly if there are errors in the data labels since it will fit them exactly. Perhaps we can help this by
                # ensembling several SVCs together each trained on a different bootstrap sample? This might add the roubustness (against label mismatches)
                # that is lacking with hard-margin SVC fitting (it is also motivated a bit by the connections between SGD and hard-margin SVC fitting, and that
                # in practice SGD works on subsamples of the data so it should be less sensitive to label misspecification).
                self.model = LinearSVC()
                # self.model = LogisticRegression(penalty='l2', random_state=0, C=1.0, solver='lbfgs', max_iter=1000)
                self.model.fit(embeddings[topic_scaling == 1], [labels[i] for i in range(len(labels)) if topic_scaling[i] == 1])
            else:
                self.model = ConstantModel(labels[0])

    def __call__(self, input, output):
        embeddings = np.hstack(adatest.embed([input, output]))
        if not hasattr(embeddings[0], "__len__"):
            return self.model.predict([embeddings])[0]
        return self.model.predict(embeddings)

class TopicMembershipModel:
    """ A model that predicts if a given test fits in a given topic.

    Note that this model only depends on the inputs not the output values for a test.
    """
    def __init__(self, topic, test_tree):
        self.topic = topic
        self.test_tree = test_tree

        # compute the weighting of samples for each test for this topic (SML: this is disabled below right now)
        # TODO: make this code shared correctly with prompt builder
        parts = topic.split("/")
        topic_scaling = np.ones(test_tree.shape[0])
        for i in range(1, len(parts)):
            prefix = "/".join(parts[:i+1])
            topic_scaling *= 1 + 9 * np.array([v.startswith(prefix) for v in test_tree["topic"]])
        topic_scaling += 1 * np.array([v == topic for v in test_tree["topic"]])

        # collect the embeddings
        # TODO: like in prompt builder we should have a max of how many samples we work with
        null_embedding = adatest.embed([""])[0]
        embeddings = []
        labels = []
        output_map = {}
        perfect_output_map = True
        for i, (id, test) in enumerate(test_tree.iterrows()):
            if test.labeler == "imputed" or test.label == "topic_marker":
                topic_scaling[i] = 0
                embeddings.append(null_embedding)
                labels.append("on_topic")
            else:
                embeddings.append(adatest.embed([test.input])[0])
                labels.append("off_topic" if test.label == "off_topic" else "on_topic")

        if len(embeddings) == 0 and len(labels) == 0: # Handle empty testframe case
            self.model = ConstantModel("Unknown")
        else:
            embeddings = np.vstack(embeddings)

            # normalize the weights
            topic_scaling /= topic_scaling.max()
            topic_scaling *= topic_scaling == 1 # TODO: this is a hack to turn off our dependenence on out-of-topic samples (unless we have zero in-topic samples)

            # check if the labels are perfectly determined by the output value
            output_map = {}
            perfect_output_map = True
            for i, (id, test) in enumerate(test_tree.iterrows()):
                if topic_scaling[i] > 0:
                    if output_map.get(test.output, test.label) != test.label:
                        perfect_output_map = False
                        break
                    output_map[test.output] = test.label
            
            # degenerate case with only one label type
            # if len(set([l for i,l in enumerate(labels) if topic_scaling[i] > 0])) == 1:
            #     self.model = ConstantModel(labels[0])
            # el
            # if perfect_output_map:
            #     self.model = OutputNearestNeighborLabelModel(embeddings, labels)
            

            # fit a linear SVM the outputs don't 
            if len(set([l for i,l in enumerate(labels) if topic_scaling[i] > 0])) > 1:
                # we are in a highly overparametrized situation, so we use a linear SVC to get "max-margin" based generalization
                # TODO: SML: It seems to me that the SVC seems to do very well as long as there are no "errors" in the data labels. But it will
                # do very poorly if there are errors in the data labels since it will fit them exactly. Perhaps we can help this by
                # ensembling several SVCs together each trained on a different bootstrap sample? This might add the roubustness (against label mismatches)
                # that is lacking with hard-margin SVC fitting (it is also motivated a bit by the connections between SGD and hard-margin SVC fitting, and that
                # in practice SGD works on subsamples of the data so it should be less sensitive to label misspecification).
                self.model = LinearSVC()
                # self.model = LogisticRegression(penalty='l2', random_state=0, C=1.0, solver='lbfgs', max_iter=1000)
                self.model.fit(embeddings[topic_scaling == 1], [labels[i] for i in range(len(labels)) if topic_scaling[i] == 1])
            else:
                self.model = ConstantModel(labels[0])

    def __call__(self, input):
        embeddings = adatest.embed([input])[0]
        if not hasattr(embeddings[0], "__len__"):
            return self.model.predict([embeddings])[0]
        return self.model.predict(embeddings)

class ChainTopicModel:
    def __init__(self, model=None):
        if model is None:
            self.base_model = RidgeClassifierCV()
        else:
            self.base_model = model
    def fit(self, X, y):
        topics = y
        max_levels = max([len(x.split('>')) for x in topics])
        self.model = sklearn.multioutput.ClassifierChain(self.base_model, order=list(range(max_levels)))
        y = [list(map(str.strip, x.split('>'))) for x in topics]
        y = np.array([x + ['-'] * (max_levels - len(x)) for x in y])
        self.encoders = [preprocessing.LabelEncoder() for _ in range(max_levels)]
        self.possible_topics = set()
        for x in topics:
            self.possible_topics.add(x)
            a = x.split(' > ')
            for i in range(1, len(a)):
                self.possible_topics.add(' > '.join(a[:i]))

        self.classes_ = list(self.possible_topics)
        new_y = np.zeros(y.shape)
        for i in range(y.shape[1]):
            self.encoders[i].fit(y[:, i])
            new_y[:, i] = self.encoders[i].transform(y[:, i])
        self.model.fit(X, new_y)
    def predict(self, X):
        y = self.model.predict(X)
        ret = []
        for i in range(y.shape[1]):
            ret.append(self.encoders[i].inverse_transform(y[:, i].astype(int)))
        y = np.array(ret).T
        ret = []
        for x in y:
            x = [z for z in x if z != '-']
            a = ' > '.join(x)
            while a not in self.possible_topics:
                x = x[:-1]
                a = ' > '.join(x)
            ret.append(a)
        return np.array(ret)

    def predict_proba(self, X):
        # This is just a fake function for now, puts 1 in the predicted class and 0 elsewhere
        y = self.predict(X)
        ret = np.zeros((len(X), len(self.classes_)))
        for i, r in enumerate(y):
            ret[i, self.classes_.index(r)] = 1
        return ret

class StandardTopicModel:
    def __init__(self, threshold=0.5):
        self.model= sklearn.linear_model.RidgeClassifierCV()
        self.threshold=threshold
        # add the missing predict_proba method to RidgeClassifierCV
        def predict_proba(self, X):
            if len(self.classes_) == 1:
                return np.ones((len(X), 1))
            d = self.decision_function(X)
            if len(self.classes_) == 2:
                probs = np.exp(d) / (np.exp(d) + np.exp(-d))
                return np.array([1 - probs, probs]).T
            probs = np.exp(d).T / np.sum(np.exp(d), axis=1)
            return probs.T
        self.model.predict_proba = predict_proba.__get__(self.model, self.model.__class__)
    def fit(self, X, y):
        self.model.fit(X, y)
    def predict_proba(self, X):
        return self.model.predict_proba(X)
    def predict(self, X):
        if self.threshold is None:
            return self.model.predict(X)
        pps = self.model.predict_proba(X)
        zero_index = list(self.model.classes_).index('Not problematic')
        ret = []
        for p in pps:
            if p[zero_index] >= self.threshold:
                ret.append(self.model.classes_[zero_index])
                continue
            else:
                best = np.argsort(p)
                if best[-1] == zero_index:
                    best = best[:-1]
                ret.append(self.model.classes_[best[-1]])
        return np.array(ret)
        # return self.model.predict(X)



from ._test_tree import TestTree
from ._test_tree_browser import TestTreeBrowser
from ._scorer import Scorer, DummyScorer, ClassifierScorer, GeneratorScorer
from ._server import serve
from ._embedding import embed
from ._model import Model
from ._topic_model import ChainTopicModel, StandardTopicModel
from . import generators

__version__ = '0.3.3'

default_generators = {
    "abstract": TestTree(r"test_trees/abstract_capabilities.csv")
}
text_embedding_model = None
image_embedding_model = None